import asyncio
import logging
import threading
import unittest
import unittest.mock
from typing import List, Any

from ._helper import async_test, AsyncMock, ExampleReadable, ExampleSubscribable, ExampleWritable, ExampleReading, \
    SimpleIntRepublisher
from shc import base


TOTALLY_RANDOM_NUMBER = 42


class TestSubscribe(unittest.TestCase):
    @async_test
    async def test_simple_subscribe(self) -> None:
        a = ExampleSubscribable(int)
        b = ExampleWritable(int)
        a.subscribe(b)
        await a.publish(TOTALLY_RANDOM_NUMBER, [self])
        b._write.assert_called_once_with(TOTALLY_RANDOM_NUMBER, [self, a])

    @async_test
    async def test_loopback_protection(self) -> None:
        a = ExampleSubscribable(int)
        b = ExampleWritable(int)
        a.subscribe(b)
        await a.publish(TOTALLY_RANDOM_NUMBER, [b])
        self.assertEqual(b._write.call_count, 0)
        await a.publish(TOTALLY_RANDOM_NUMBER, [b, self])
        self.assertEqual(b._write.call_count, 0)

    @async_test
    async def test_error_handling(self) -> None:
        a = ExampleSubscribable(int)
        b = ExampleWritable(int)
        b._write.side_effect = RuntimeError("Really unexpected error in _write")
        a.subscribe(b)
        with self.assertLogs(level=logging.ERROR) as ctx:
            await a.publish(TOTALLY_RANDOM_NUMBER, [self])
        self.assertIn("unexpected error in _write", "\n".join(ctx.output))  # type: ignore

    @async_test
    async def test_type_conversion(self) -> None:
        a = ExampleSubscribable(int)
        b = ExampleWritable(float)

        with self.assertRaises(TypeError):
            a.subscribe(b)
        a.subscribe(b, convert=True)
        await a.publish(TOTALLY_RANDOM_NUMBER, [self])
        b._write.assert_called_once()
        self.assertEqual(float(TOTALLY_RANDOM_NUMBER), b._write.call_args[0][0])
        self.assertIsInstance(b._write.call_args[0][0], float)

        c = ExampleWritable(list)
        with self.assertRaises(TypeError):
            a.subscribe(c)

    @async_test
    async def test_explicit_conversion(self) -> None:
        a = ExampleSubscribable(int)
        b = ExampleWritable(float)

        converter = unittest.mock.Mock(side_effect=(56.5,))

        a.subscribe(b, convert=converter)
        await a.publish(TOTALLY_RANDOM_NUMBER, [self])
        converter.assert_called_once_with(TOTALLY_RANDOM_NUMBER)
        b._write.assert_called_once_with(56.5, unittest.mock.ANY)


class TestHandler(unittest.TestCase):
    @async_test
    async def test_basic_trigger(self) -> None:
        a = ExampleSubscribable(int)
        handler = AsyncMock()
        handler.__name__ = "handler_mock"
        wrapped_handler = base.handler()(handler)
        a.trigger(wrapped_handler)
        await a.publish(TOTALLY_RANDOM_NUMBER, [self])
        await asyncio.sleep(0.01)
        handler.assert_called_once_with(TOTALLY_RANDOM_NUMBER, [self, a])

    @async_test
    async def test_magic_origin_passing(self) -> None:
        a = ExampleSubscribable(int)
        b = ExampleWritable(int)

        @a.trigger
        @base.handler()
        async def test_handler(value, _origin) -> None:
            await b.write(value)

        await a.publish(TOTALLY_RANDOM_NUMBER, [self])
        await asyncio.sleep(0.01)
        b._write.assert_called_once_with(TOTALLY_RANDOM_NUMBER, [self, a, test_handler])

    @async_test
    async def test_reset_origin(self) -> None:
        a = ExampleSubscribable(int)
        b = ExampleWritable(int)

        @a.trigger
        @base.handler(reset_origin=True)
        async def test_handler(value, _origin) -> None:
            await b.write(value)

        await a.publish(TOTALLY_RANDOM_NUMBER, [self])
        await asyncio.sleep(0.01)
        b._write.assert_called_once_with(TOTALLY_RANDOM_NUMBER, [test_handler])

    @async_test
    async def test_no_recursion(self) -> None:
        a = SimpleIntRepublisher()
        self.call_counter = 0

        @a.trigger
        @base.handler()
        async def a_test_handler(value, _origin):
            self.call_counter += 1
            if value > TOTALLY_RANDOM_NUMBER:
                return
            await a.write(value + 1)

        await a_test_handler(TOTALLY_RANDOM_NUMBER, [])
        self.assertEqual(1, self.call_counter)

    @async_test
    async def test_allow_recursion(self) -> None:
        a = SimpleIntRepublisher()
        self.call_counter = 0

        @a.trigger
        @base.handler(allow_recursion=True)
        async def another_test_handler(value, _origin):
            self.call_counter += 1
            if value > TOTALLY_RANDOM_NUMBER:
                return
            await a.write(value + 1)

        await another_test_handler(TOTALLY_RANDOM_NUMBER, [])
        await asyncio.sleep(0.01)
        self.assertEqual(2, self.call_counter)

    @async_test
    async def test_error_handling(self) -> None:
        a = ExampleSubscribable(int)
        mock = AsyncMock(side_effect=RuntimeError("Really unexpected error in _write"))

        @a.trigger
        @base.handler()
        async def test_handler(value, _origin) -> None:
            await mock()

        with self.assertLogs(level=logging.ERROR) as ctx:
            await a.publish(TOTALLY_RANDOM_NUMBER, [self])
            await asyncio.sleep(0.01)
        self.assertIn("unexpected error in _write", "\n".join(ctx.output))  # type: ignore


class TestBlockingHandler(unittest.TestCase):
    @async_test
    async def test_basic_trigger(self) -> None:
        a = ExampleSubscribable(int)
        thread_id_container = []

        def save_thread(*args):
            thread_id_container.append(threading.get_ident())

        handler = unittest.mock.Mock(side_effect=save_thread)
        handler.__name__ = "handler_mock"
        wrapped_handler = base.blocking_handler()(handler)
        a.trigger(wrapped_handler)
        await a.publish(TOTALLY_RANDOM_NUMBER, [self])
        await asyncio.sleep(0.01)
        handler.assert_called_once_with(TOTALLY_RANDOM_NUMBER, [self, a])
        self.assertIsInstance(thread_id_container[0], int)
        self.assertNotEqual(thread_id_container[0], threading.get_ident())

    @async_test
    async def test_magic_origin_receiving(self) -> None:
        a = ExampleSubscribable(int)
        mock = unittest.mock.Mock()

        @a.trigger
        @base.handler()
        async def test_handler(value, _origin) -> None:
            await blocking_test_handler(value)  # type: ignore  # MyPy doesn't get that @blocking_handler() changes args

        @base.blocking_handler()
        def blocking_test_handler(value, _origin):
            mock(value, _origin)

        await a.publish(TOTALLY_RANDOM_NUMBER, [self])
        await asyncio.sleep(0.01)
        mock.assert_called_once_with(TOTALLY_RANDOM_NUMBER, [self, a, test_handler])


class TestReading(unittest.TestCase):
    @async_test
    async def test_simple_reading(self) -> None:
        a = ExampleReadable(int, TOTALLY_RANDOM_NUMBER)
        b = ExampleReading(int, False)
        b.set_provider(a)
        result = await b.do_read()
        self.assertEqual(result, TOTALLY_RANDOM_NUMBER)

    @async_test
    async def test_type_conversion(self) -> None:
        a = ExampleReadable(int, TOTALLY_RANDOM_NUMBER)
        b = ExampleReading(float, False)
        with self.assertRaises(TypeError):
            b.set_provider(a)

        b.set_provider(a, convert=True)
        result = await b.do_read()
        self.assertEqual(result, float(TOTALLY_RANDOM_NUMBER))
        self.assertIsInstance(result, float)

        c = ExampleReading(list, False)
        with self.assertRaises(TypeError):
            c.set_provider(a, convert=True)

    @async_test
    async def test_explicit_conversion(self) -> None:
        a = ExampleReadable(int, TOTALLY_RANDOM_NUMBER)
        b = ExampleReading(float, False)

        converter = unittest.mock.Mock(side_effect=(56.5,))

        b.set_provider(a, convert=converter)

        result = await b.do_read()
        converter.assert_called_once_with(TOTALLY_RANDOM_NUMBER)
        self.assertEqual(result, 56.5)


class DummyIntReadSubscribable(base.Readable[int], base.Subscribable[int]):
    type = int

    def __init__(self):
        super().__init__()

    async def read(self) -> int:
        pass


class DummyFloatReadingWritable(base.Reading[float], base.Writable[float]):
    is_reading_optional = False
    type = float

    def __init__(self):
        super().__init__()

    async def _write(self, value: float, origin: List[Any]):
        pass


class DummyIntWrapper(base.ConnectableWrapper[int]):
    type = int
    connect: unittest.mock.Mock  # required to let MyPy know that we can use Mock's methods

    def __init__(self):
        self.connect = unittest.mock.Mock()

    def connect(self, *args, **kwargs) -> "DummyIntWrapper": ...  # type: ignore


class TestConnecting(unittest.TestCase):
    def test_subscribing(self) -> None:
        a = ExampleSubscribable(int)
        b = ExampleWritable(int)

        with unittest.mock.patch.object(a, 'subscribe') as mock_subscribe:
            a.connect(b, receive=False)  # `read` should not have an effect here
            mock_subscribe.assert_called_once_with(b, convert=False)

            mock_subscribe.reset_mock()
            b.connect(a, send=False)  # `provide` should not have an effect here
            mock_subscribe.assert_called_once_with(b, convert=False)

            # Explicit override
            mock_subscribe.reset_mock()
            a.connect(b, send=False)
            mock_subscribe.assert_not_called()

            b.connect(a, receive=False)
            mock_subscribe.reset_mock()
            mock_subscribe.assert_not_called()

        # Exceptions
        with self.assertRaises(TypeError):
            a.connect(b, receive=True)
        with self.assertRaises(TypeError):
            b.connect(a, send=True)

    def test_mandatory_reading(self) -> None:
        a = ExampleReadable(int, TOTALLY_RANDOM_NUMBER)
        b = ExampleReading(int, False)

        with unittest.mock.patch.object(b, 'set_provider') as mock_set_provider:
            # Since Reading is mandatory, a should be registered with b by default
            a.connect(b, read=False)  # `read=False` should not have an effect here
            mock_set_provider.assert_called_once_with(a, convert=False)

            mock_set_provider.reset_mock()
            b.connect(a, provide=False)  # `provide=False` should not have an effect here
            mock_set_provider.assert_called_once_with(a, convert=False)

            # Explicit override
            mock_set_provider.reset_mock()
            a.connect(b, provide=False)
            mock_set_provider.assert_not_called()

            mock_set_provider.reset_mock()
            b.connect(a, read=False)
            mock_set_provider.assert_not_called()

        # Exceptions
        with self.assertRaises(TypeError):
            a.connect(b, read=True)
        with self.assertRaises(TypeError):
            b.connect(a, provide=True)

    def test_optional_reading(self) -> None:
        a = ExampleReadable(int, TOTALLY_RANDOM_NUMBER)
        b = ExampleReading(int, True)

        with unittest.mock.patch.object(b, 'set_provider') as mock_set_provider:
            a.connect(b)
            mock_set_provider.assert_not_called()

            mock_set_provider.reset_mock()
            b.connect(a, read=True)
            mock_set_provider.assert_called_once_with(a, convert=False)

    # TODO test None return values of _from_provider

    def test_type_conversion(self) -> None:
        a = DummyIntReadSubscribable()
        b = DummyFloatReadingWritable()

        with self.assertRaises(TypeError):
            a.connect(b)

        with self.assertRaises(TypeError):
            b.connect(a)

        with unittest.mock.patch.object(a, 'subscribe') as mock_subscribe,\
                unittest.mock.patch.object(b, 'set_provider') as mock_set_provider:
            a.connect(b, convert=True)
            mock_subscribe.assert_called_once_with(b, convert=True)
            mock_set_provider.assert_called_once_with(a, convert=True)

            mock_subscribe.reset_mock()
            mock_set_provider.reset_mock()
            b.connect(a, convert=True)
            mock_subscribe.assert_called_once_with(b, convert=True)
            mock_set_provider.assert_called_once_with(a, convert=True)

    def test_explict_conversion(self) -> None:
        a = DummyIntReadSubscribable()
        b = DummyFloatReadingWritable()

        def a2b(x: int) -> float:
            return float(x/255)

        def b2a(x: float) -> int:
            return round(x*255)

        with unittest.mock.patch.object(a, 'subscribe') as mock_subscribe,\
                unittest.mock.patch.object(b, 'set_provider') as mock_set_provider:
            a.connect(b, convert=(a2b, b2a))
            mock_subscribe.assert_called_once_with(b, convert=a2b)
            mock_set_provider.assert_called_once_with(a, convert=a2b)

            mock_subscribe.reset_mock()
            mock_set_provider.reset_mock()
            b.connect(a, convert=(b2a, a2b))
            mock_subscribe.assert_called_once_with(b, convert=a2b)
            mock_set_provider.assert_called_once_with(a, convert=a2b)

    def test_connectable_wrapper(self) -> None:
        a = DummyIntReadSubscribable()
        b = DummyIntWrapper()

        def a2b(x: int) -> int:
            return x*2

        def b2a(x: int) -> int:
            return x//2

        a.connect(b, read=False, provide=True, convert=True)
        b.connect.assert_called_once_with(a, send=None, receive=None, read=True, provide=False, convert=True)

        b.connect.reset_mock()
        a.connect(b, send=False, receive=True, convert=(a2b, b2a))
        b.connect.assert_called_once_with(a, send=True, receive=False, read=None, provide=None, convert=(b2a, a2b))



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import asyncio
import datetime
import enum
import logging
import math
from typing import List, Any, Dict, Tuple, Optional, Set, Generic, NamedTuple

import knxdclient
from ._helper import SupervisedClientInterface

from .. import datatypes
from ..base import Writable, Subscribable, Reading, T
from ..conversion import register_converter
from ..datatypes import FadeStep

KNXGAD = knxdclient.GroupAddress

logger = logging.getLogger(__name__)


class KNXHVACMode(enum.Enum):
    """
    Python enum representation of the KNX datapoint type 20.102 "DPT_HVACMode", a 8-bit enum of heating/ventilation/AC
    operating modes.

    The value mapping corresponds to KNX' native value encoding of this datatype.
    """
    AUTO = 0
    COMFORT = 1
    STANDBY = 2
    ECONOMY = 3
    BUILDING_PROTECTION = 4


class KNXUpDown(enum.Enum):
    """
    Python enum representation of the KNX datapoint type 1.008 "DPT_UpDown", a 1-bit value for controlling blinds etc.

    Values of this type can also be used as bool values, using the native KNX value mapping (to datapoint type 1.001).
    """
    UP = False
    DOWN = True

    def __bool__(self):
        return self.value


class KNXControlDimming(NamedTuple):
    """
    Python NamedTuple representation of the KNX datapoint type 3.007 "DPT_Control_Dimming" or 3.008
    "DPT_Control_Blinds".

    :param increase: True: Increase dimmer brightness / lower blinds; False: Decrese dimmer brightness / raise blinds
    :param step_exponent: 0: Break dimmming action; 1-7 define step size = 2^(1-stepcode)
    """
    increase: bool
    step_exponent: int

    @property
    def step(self) -> FadeStep:
        if not self.step_exponent:
            return FadeStep(0.0)
        return FadeStep((1 if self.increase else -1) * 2.0 ** (1 - self.step_exponent))

    @classmethod
    def from_step(cls, value: FadeStep) -> "KNXControlDimming":
        increase = True
        val: float = value
        if val < 0:
            increase = False
            val = -val
        # Rounding the logarithm doesn't give us the nearest exponent of two of the original value, but it should be
        # okay in practice.
        return cls(increase, 0 if val == 0.0 else min(7, 1 - round(math.log2(val))))


register_converter(KNXHVACMode, int, lambda v: v.value)
register_converter(int, KNXHVACMode, lambda v: KNXHVACMode(v))
register_converter(KNXUpDown, bool, lambda v: v.value)
register_converter(bool, KNXUpDown, lambda v: KNXUpDown(v))
register_converter(KNXControlDimming, FadeStep, lambda v: v.step)
register_converter(KNXControlDimming, float, lambda v: v.step)
register_converter(FadeStep, KNXControlDimming, lambda v: KNXControlDimming.from_step(v))
register_converter(float, KNXControlDimming, lambda v: KNXControlDimming.from_step(FadeStep(min(1.0, max(-1.0, v)))))
register_converter(datetime.datetime, knxdclient.KNXTime, knxdclient.KNXTime.from_datetime)


KNXDPTs: Dict[str, Tuple[type, knxdclient.KNXDPT]] = {
    '1': (bool, knxdclient.KNXDPT.BOOLEAN),
    '1.008': (KNXUpDown, knxdclient.KNXDPT.BOOLEAN),
    '3': (KNXControlDimming, knxdclient.KNXDPT.BOOLEAN_UINT3),
    '4': (str, knxdclient.KNXDPT.CHAR),
    '5': (int, knxdclient.KNXDPT.UINT8),
    '5.001': (datatypes.RangeUInt8, knxdclient.KNXDPT.UINT8),
    '5.003': (datatypes.AngleUInt8, knxdclient.KNXDPT.UINT8),
    '5.004': (datatypes.RangeInt0To100, knxdclient.KNXDPT.UINT8),
    '6': (int, knxdclient.KNXDPT.INT8),
    '7': (int, knxdclient.KNXDPT.UINT16),
    '8': (int, knxdclient.KNXDPT.INT16),
    '9': (float, knxdclient.KNXDPT.FLOAT16),
    '10': (knxdclient.KNXTime, knxdclient.KNXDPT.TIME),
    '11': (datetime.date, knxdclient.KNXDPT.DATE),
    '12': (int, knxdclient.KNXDPT.UINT32),
    '13': (int, knxdclient.KNXDPT.INT32),
    '14': (float, knxdclient.KNXDPT.FLOAT32),
    '16': (str, knxdclient.KNXDPT.STRING),
    '17': (int, knxdclient.KNXDPT.SCENE_NUMBER),
    '19': (datetime.datetime, knxdclient.KNXDPT.DATE_TIME),
    '20.102': (KNXHVACMode, knxdclient.KNXDPT.ENUM8),
}


class KNXConnector(SupervisedClientInterface):
    """
    SHC interface for connecting with a KNX home automation bus via KNX deamon (KNXD).

    The interface allows to interact bidirectional with KNX group addresses (i.e. send and receive KNX *group write*
    telegrams). For this purpose a *Connectable* object can be created for each group address, using the :meth:`group`
    method.

    The connection to the KNX bus is established by using KNXDs native client protocol (not via KNX over UDP protocol),
    either via TCP port or via UNIX domain socket. Thus, KNXD must be started with either the `-i` or `-u` option
    (or be run by systemd with an appropriate config for taking care of this). By default, the `KNXConnector` tries to
    connect to KNXDs default TCP port 6720 at localhost. The parameters ``host``, ``port``, and ``sock`` can be used to
    specify another host/port or connect via a local UNIX domain socket instead.

    :param host: Hostname for connecting to KNXD via TCP. Defaults to 'localhost'
    :param port: TCP port where KNXD is listening for client connections at the specified `host`. Defaults to 6720.
    :param sock: Path to the KNXD UNIX domain socket. If given, it is used instead of the TCP connection to host/port.
    :param auto_reconnect: If True (default), the interface tries to reconnect automatically with exponential backoff
        (5 * 1.25^n seconds sleep), when connection to KNXD is lost. Otherwise, the complete SHC system is shut down
        on connection errors.
    :param read_init_after_reconnect: If True (default), the *group read* telegrams for initialization (according to the
        `init` parameter of :meth:`group`) are resent when reconnecting after a connection loss to compensate for
        possibly missed value updates on the KNX bus.
    :param failsafe_start: If True, the KNXD client allows SHC to start up, even if the connection to KNXD can not be
        established in the first try. The connection is retried in background with exponential backoff (see
        `auto_reconnect` option). Otherwise (default), the first connection attempt on startup is not retried and will
        shutdown the SHC application on failure, even if `auto_reconnect` is True.
    """
    def __init__(self, host: str = 'localhost', port: int = 6720, sock: Optional[str] = None,
                 auto_reconnect: bool = True, read_init_after_reconnect: bool = True, failsafe_start: bool = False):
        super().__init__(auto_reconnect, failsafe_start)
        self.backoff_base = 5
        self.host = host
        self.port = port
        self.sock = sock
        self.read_init_after_reconnect = read_init_after_reconnect
        self.groups: Dict[KNXGAD, KNXGroupVar] = {}
        self.knx = knxdclient.KNXDConnection()
        self.knx.set_group_apdu_handler(self._dispatch_telegram)
        self.init_request_groups: Set[KNXGAD] = set()
        self._first_connect = True

    async def _connect(self) -> None:
        await self.knx.connect(self.host, self.port, self.sock)

    async def _subscribe(self) -> None:
        await self.knx.open_group_socket()
        if self._first_connect or self.read_init_after_reconnect:
            await self._send_init_requests()
        self._first_connect = False

    async def _run(self):
        self._running.set()
        await self.knx.run()

    async def _disconnect(self) -> None:
        await self.knx.stop()

    def group(self, addr: KNXGAD, dpt: str, init: bool = False) -> "KNXGroupVar":
        """
        Create a *Connectable* object for sending and receiving KNX telegrams for a given group address.

        The returned object is *Subscribable* for receiving updates (group write and group response telegrams) from the
        KNX system and *Writable* to send a new value to the KNX system. It is also optionally *Reading*. If a
        `default_provider` is set (e.g. via the read/provide parameter of
        :meth:`connect <shc.base.Connectable.connect>`), this `KNXConnector` actively responds to *group read* telegrams
        from the KNX system by sending a *group response* with the read value.

        To ensure correct data encoding, the KNX datapoint type of the group address must be specified. It **must** be
        equal to the datapoint type of other KNX devices' datapoints which are connected to this group address. The
        returned *Connectable's* `type` is derived from the KNX datapoint type. The following KNX datapoint types (DPT)
        are supported:

        +----------+---------------------------------------+
        | KNX DPT  | Python `type`                         |
        +==========+=======================================+
        | '1'      | :class:`bool`                         |
        +----------+---------------------------------------+
        | '1.008'  | :class:`KNXUpDown`                    |
        +----------+---------------------------------------+
        | '3'      | :class:`KNXControlDimming`            |
        +----------+---------------------------------------+
        | '4'      | :class:`str`                          |
        +----------+---------------------------------------+
        | '5'      | :class:`int`                          |
        +----------+---------------------------------------+
        | '5.001'  | :class:`shc.datatypes.RangeUInt8`     |
        +----------+---------------------------------------+
        | '5.003'  | :class:`shc.datatypes.AngleUInt8`     |
        +----------+---------------------------------------+
        | '5.004'  | :class:`shc.datatypes.RangeInt0To100` |
        +----------+---------------------------------------+
        | '6'      | :class:`int`                          |
        +----------+---------------------------------------+
        | '7'      | :class:`int`                          |
        +----------+---------------------------------------+
        | '8'      | :class:`int`                          |
        +----------+---------------------------------------+
        | '9'      | :class:`float`                        |
        +----------+---------------------------------------+
        | '10'     | :class:`knxdclient.KNXTime`           |
        +----------+---------------------------------------+
        | '11'     | :class:`datetime.date`                |
        +----------+---------------------------------------+
        | '12'     | :class:`int`                          |
        +----------+---------------------------------------+
        | '13'     | :class:`int`                          |
        +----------+---------------------------------------+
        | '14'     | :class:`float`                        |
        +----------+---------------------------------------+
        | '16'     | :class:`str`                          |
        +----------+---------------------------------------+
        | '17'     | :class:`int`                          |
        +----------+---------------------------------------+
        | '19'     | :class:`datetime.datetime`            |
        +----------+---------------------------------------+
        | '20.102' | :class:`KNXHVACMode`                  |
        +----------+---------------------------------------+

        When `group` is called multiple times with the same group address, a reference to the **same** *Connectable*
        object is returned. This ensures, that dispatching of incoming messages and local feedback (see below) always
        work correctly and checks on the `origin` of a new value don't behave unexpectedly. However, the datapoint type
        given in all calls for the same group address must match. Otherwise, a `ValueError` is raised.

        The *Connectable* object for each group address features an internal local feedback. This means, that every
        new value *written* to the object is being published to all other local subscribers, **after** being transmitted
        to the KNX system. Thus, the KNX group address behaves "bus-like", for the connected objects within SHC: When
        one connected object sends a new value to the KNX bus, its received by all KNX devices and all other connected
        objects as well, which is important for central functions etc.

        :param addr: The KNX group address to connect to, represented as a :class:`KNXGAD` object
        :param dpt: The KNX datapoint type (DPT) number as string according to the table above
        :param init: If True, the interface will send a *group read* telegram to this group address after SHC's startup.
            This can be used to initialize subscribed SHC variables with the current value from the KNX system, if
            there's a KNX device responding to read requests for this group address (i.e. which has the *read* flag set
            on the relevant datapoint).
        :return: The *Connectable* object representing the group address
        :raise ValueError: If `group` has been called before with the same group address but a different datapoint type
        """
        if addr in self.groups:
            group_var = self.groups[addr]
            if group_var.dpt != dpt:
                raise ValueError("KNX Datapoint Type conflict: Group Variable {} has been created with type {} before"
                                 .format(group_var.addr, group_var.dpt))
        else:
            group_var = KNXGroupVar(self, addr, dpt)
            self.groups[addr] = group_var
        if init:
            self.init_request_groups.add(addr)
        return group_var

    async def _send_init_requests(self):
        await asyncio.gather(*(self.knx.group_write(addr, knxdclient.KNXDAPDUType.READ, 0)
                               for addr in self.init_request_groups))

    def _dispatch_telegram(self, packet: knxdclient.ReceivedGroupAPDU) -> None:
        if packet.payload.type is knxdclient.KNXDAPDUType.READ:
            if packet.dst in self.groups:
                asyncio.create_task(self._respond_read_request(self.groups[packet.dst]))
        else:
            try:
                group_var = self.groups[packet.dst]
            except KeyError:
                logging.debug("No KNX Group Variable for Addr %s registered", packet.dst)
                return
            group_var.update_from_bus(packet.payload.value, [packet.src])

    async def _respond_read_request(self, group_var: "KNXGroupVar") -> None:
        try:
            encoded_data = await group_var.read_from_bus()
            if encoded_data is not None:
                await self.knx.group_write(group_var.addr, knxdclient.KNXDAPDUType.RESPONSE, encoded_data)
        except Exception as e:
            logger.warning("Error while responding to KNX group read request for %s:", group_var.addr, exc_info=e)

    async def send(self, addr: knxdclient.GroupAddress, encoded_data: knxdclient.EncodedData):
        await self.knx.group_write(addr, knxdclient.KNXDAPDUType.WRITE, encoded_data)
        await asyncio.sleep(0.15)  # wait a bit longer to be sure, that the telegram has been sent to the bus

    def __repr__(self) -> str:
        return "{}(host={}, port={}, sock={})".format(self.__class__.__name__, self.host, self.port, self.sock)


class KNXGroupVar(Subscribable[T], Writable[T], Reading[T], Generic[T]):
    _stateful_publishing = True

    def __init__(self, connector: KNXConnector, addr: KNXGAD, dpt: str):
        if dpt not in KNXDPTs:
            raise ValueError("KNX Datapoint Type {} is not supported".format(dpt))
        self.type = KNXDPTs[dpt][0]
        super().__init__()
        self.dpt = dpt
        self.knx_major_dpt = KNXDPTs[dpt][1]
        self.connector = connector
        self.addr = addr

    def update_from_bus(self, data: knxdclient.EncodedData, origin: List[Any]) -> None:
        value: T = knxdclient.decode_value(data, self.knx_major_dpt)
        if type(value) is not self.type:
            value = self.type(*value) if issubclass(self.type, tuple) else self.type(value)  # type: ignore
        logger.debug("Got new value %s for KNX Group variable %s from bus", value, self.addr)
        self._publish(value, origin)

    async def read_from_bus(self) -> Optional[knxdclient.EncodedData]:
        value = await self._from_provider()
        if value is not None:
            return knxdclient.encode_value(value, self.knx_major_dpt)
        return None

    async def _write(self, value: T, origin: List[Any]) -> None:
        self._publish(value, origin)
        encoded_data = knxdclient.encode_value(value, self.knx_major_dpt)
        await self.connector.send(self.addr, encoded_data)

    def __repr__(self) -> str:
        return "{}(GAD={})".format(self.__class__.__name__, self.addr)



import asyncio
import datetime
import unittest
import unittest.mock
from unittest.mock import Mock
from typing import Optional, List

import shc.base
from shc import timer, base, datatypes
from ._helper import ClockMock, async_test, ExampleSubscribable, AsyncMock, ExampleWritable, ExampleReadable


class LogarithmicSleepTest(unittest.TestCase):
    @async_test
    async def test_long_sleep(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 0, 0)) as clock:
            await timer._logarithmic_sleep(datetime.datetime(2020, 1, 17, 18, 5, 13).astimezone())
            self.assertEqual(datetime.datetime(2020, 1, 17, 18, 5, 13), clock.now())

    @async_test
    async def test_short_sleep(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 0, 0)) as clock:
            await timer._logarithmic_sleep(datetime.datetime(2020, 1, 1, 15, 0, 1, 17600).astimezone())
            self.assertEqual(datetime.datetime(2020, 1, 1, 15, 0, 1, 17600), clock.now())

    @async_test
    async def test_sleep_overshoot(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 0, 0), datetime.timedelta(microseconds=27600)) as clock:
            await timer._logarithmic_sleep(datetime.datetime(2020, 1, 1, 18, 37, 10).astimezone())
            miss_by = clock.now() - datetime.datetime(2020, 1, 1, 18, 37, 10)
            self.assertGreaterEqual(miss_by, datetime.timedelta(0))
            self.assertLess(miss_by, 2 * datetime.timedelta(microseconds=27600))

    @async_test
    async def test_negative_delay(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 0, 0)) as clock:
            await timer._logarithmic_sleep(datetime.datetime(2020, 1, 1, 14, 59, 25).astimezone())
            self.assertEqual(datetime.datetime(2020, 1, 1, 15, 0, 0), clock.now())


class AbstractTimerTest(unittest.TestCase):
    class TestTimer(timer._AbstractScheduleTimer):
        def __init__(self, times: List[datetime.datetime]):
            super().__init__()
            self.times = times
            self.index = -1

        def _next_execution(self) -> Optional[datetime.datetime]:
            self.index += 1
            if self.index >= len(self.times):
                return None
            return self.times[self.index]

    @async_test
    async def test_run(self) -> None:
        clock_mock = ClockMock(datetime.datetime(2020, 8, 20, 21, 8, 0), actual_sleep=0.01)
        expected_events = [datetime.datetime(2020, 8, 20, 21, 8, 2).astimezone(),
                           datetime.datetime(2020, 8, 20, 22, 0, 0).astimezone()]
        events = []

        def store_time(*args, **kwargs):
            events.append(clock_mock.now().astimezone())

        t = self.TestTimer(expected_events)
        with unittest.mock.patch('shc.timer._AbstractScheduleTimer._publish', new=store_time):
            with clock_mock:
                await t.run()
        await asyncio.sleep(0.01)  # Allow async tasks to run to make sure all _publish tasks have been executed
        self.assertListEqual(expected_events, events)

        assert(t.last_execution is not None)
        self.assertAlmostEqual(t.last_execution,
                               expected_events[-1],
                               delta=datetime.timedelta(seconds=1))


class TimerDecoratorText(unittest.TestCase):
    def test_decorators(self) -> None:
        for decorator, timer_class, attr in ((timer.every, timer.Every, {'delta': datetime.timedelta(seconds=5)}),
                                             (timer.once, timer.Once, {'offset': datetime.timedelta(seconds=7)}),
                                             (timer.at, timer.At, {'random': datetime.timedelta(seconds=20)}),):
            with self.subTest(decorator=decorator.__name__):
                async def my_function(_val, _origin):
                    return

                with unittest.mock.patch('shc.base.Subscribable.trigger',
                                         autospec=True, side_effect=lambda s, t: t) as trigger_mock:
                    returned = decorator(**attr)(my_function)

                trigger_mock.assert_called_once()
                self.assertIsInstance(trigger_mock.call_args[0][0], timer_class)
                for attr_name, attr_value in attr.items():
                    self.assertEqual(getattr(trigger_mock.call_args[0][0], attr_name), attr_value)
                self.assertIs(trigger_mock.call_args[0][1], my_function)
                self.assertIs(trigger_mock.call_args[0][1], returned)


class EveryTimerTest(unittest.TestCase):
    def test_unaligned(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
            every_timer = timer.Every(datetime.timedelta(minutes=5), align=False)
            next_execution = every_timer._next_execution()
            assert(next_execution is not None)
            self.assertAlmostEqual(clock.now().astimezone(), next_execution,
                                   delta=datetime.timedelta(seconds=1))
            every_timer.last_execution = clock.now().astimezone()
            next_execution = every_timer._next_execution()
            assert(next_execution is not None)
            self.assertAlmostEqual(clock.now().astimezone() + datetime.timedelta(minutes=5), next_execution,
                                   delta=datetime.timedelta(seconds=1))
            clock.sleep(5 * 60)
            every_timer.last_execution = clock.now().astimezone()
            next_execution = every_timer._next_execution()
            assert(next_execution is not None)
            self.assertAlmostEqual(clock.now().astimezone() + datetime.timedelta(minutes=5), next_execution,
                                   delta=datetime.timedelta(seconds=1))

    def test_unaligned_random(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
            every_timer = timer.Every(datetime.timedelta(minutes=5), align=False, random=datetime.timedelta(seconds=20))
            self.assertGreaterEqual(every_timer._next_execution(),
                                    clock.now().astimezone() - datetime.timedelta(seconds=20))
            self.assertLessEqual(every_timer._next_execution(),
                                 clock.now().astimezone() + datetime.timedelta(seconds=20))
            every_timer.last_execution = clock.now().astimezone()
            self.assertGreaterEqual(every_timer._next_execution(),
                                    clock.now().astimezone() + datetime.timedelta(minutes=5, seconds=-20))
            self.assertLessEqual(every_timer._next_execution(),
                                 clock.now().astimezone() + datetime.timedelta(minutes=5, seconds=20))
            clock.sleep(5 * 60 + 20)
            every_timer.last_execution = clock.now().astimezone()
            self.assertGreaterEqual(every_timer._next_execution(),
                                    clock.now().astimezone() + datetime.timedelta(minutes=5, seconds=-20))
            self.assertLessEqual(every_timer._next_execution(),
                                 clock.now().astimezone() + datetime.timedelta(minutes=5, seconds=20))

    def test_aligned(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
            every_timer = timer.Every(datetime.timedelta(minutes=5), align=True)
            base = every_timer._next_execution()
            assert(base is not None)
            self.assertGreaterEqual(base - clock.now().astimezone(), datetime.timedelta(0))
            self.assertLessEqual(base - clock.now().astimezone(), datetime.timedelta(minutes=5))

            clock.current_time = base + datetime.timedelta(microseconds=1)  # We slept till the first execution
            next_execution = every_timer._next_execution()
            assert(next_execution is not None)
            self.assertAlmostEqual(clock.now().astimezone() + datetime.timedelta(minutes=5),
                                   next_execution,
                                   delta=datetime.timedelta(seconds=1))

        # A new timer (after a restart 5 minutes later) should give us exactly base + 5 minutes as the first execution
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17) + datetime.timedelta(minutes=5)) as clock:
            every_timer = timer.Every(datetime.timedelta(minutes=5), align=True)
            next_execution = every_timer._next_execution()
            assert(next_execution is not None)
            self.assertAlmostEqual(base + datetime.timedelta(minutes=5),
                                   next_execution,
                                   delta=datetime.timedelta(seconds=1))


class OnceTimerTest(unittest.TestCase):
    def test_immediate(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
            once_timer = timer.Once()
            next_execution = once_timer._next_execution()
            assert(next_execution is not None)
            self.assertAlmostEqual(clock.now().astimezone(), next_execution,
                                   delta=datetime.timedelta(seconds=1))
            clock.sleep(1)
            once_timer.last_execution = clock.now().astimezone()
            self.assertIsNone(once_timer._next_execution())

    def test_offset(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
            once_timer = timer.Once(datetime.timedelta(hours=1))
            next_execution = once_timer._next_execution()
            assert(next_execution is not None)
            self.assertAlmostEqual(clock.now().astimezone() + datetime.timedelta(hours=1), next_execution,
                                   delta=datetime.timedelta(seconds=1))
            clock.sleep(60*60)
            once_timer.last_execution = clock.now().astimezone()
            self.assertIsNone(once_timer._next_execution())

    def test_offset_random(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
            once_timer = timer.Once(datetime.timedelta(hours=1), random=datetime.timedelta(seconds=20))
            next_execution = once_timer._next_execution()
            self.assertGreaterEqual(next_execution,
                                    clock.now().astimezone() + datetime.timedelta(hours=1, seconds=-20))
            self.assertLessEqual(next_execution,
                                 clock.now().astimezone() + datetime.timedelta(hours=1, seconds=20))
            clock.sleep(60*60)
            once_timer.last_execution = clock.now().astimezone()
            self.assertIsNone(once_timer._next_execution())


class AtTimerTest(unittest.TestCase):
    def _assert_datetime(self, expected: datetime.datetime, actual: Optional[datetime.datetime]) -> None:
        assert(actual is not None)
        self.assertAlmostEqual(expected.astimezone(), actual, delta=datetime.timedelta(seconds=.1))

    def test_simple_next(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
            once_timer = timer.At(hour=15, minute=7, second=17, millis=200)
            self._assert_datetime(datetime.datetime(2020, 1, 1, 15, 7, 17, 200000), once_timer._next_execution())
            once_timer = timer.At(hour=15, minute=7, second=25)
            self._assert_datetime(datetime.datetime(2020, 1, 1, 15, 7, 25), once_timer._next_execution())
            once_timer = timer.At(hour=15, minute=12)
            self._assert_datetime(datetime.datetime(2020, 1, 1, 15, 12, 0), once_timer._next_execution())
            once_timer = timer.At(hour=16)
            self._assert_datetime(datetime.datetime(2020, 1, 1, 16, 0, 0), once_timer._next_execution())
            once_timer = timer.At(day=19)
            self._assert_datetime(datetime.datetime(2020, 1, 19, 0, 0, 0), once_timer._next_execution())
            once_timer = timer.At(month=8)
            self._assert_datetime(datetime.datetime(2020, 8, 1, 0, 0, 0), once_timer._next_execution())
            once_timer = timer.At(year=2021)
            self._assert_datetime(datetime.datetime(2021, 1, 1, 0, 0, 0), once_timer._next_execution())
            once_timer = timer.At(weekday=5)
            self._assert_datetime(datetime.datetime(2020, 1, 3, 0, 0, 0), once_timer._next_execution())
            once_timer = timer.At(weeknum=15)
            self._assert_datetime(datetime.datetime(2020, 4, 6, 0, 0, 0), once_timer._next_execution())

    def test_spec_forms(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
            once_timer = timer.At(hour=timer.EveryNth(2))
            self._assert_datetime(datetime.datetime(2020, 1, 1, 16, 0, 0), once_timer._next_execution())
            once_timer = timer.At(hour=timer.EveryNth(6))
            self._assert_datetime(datetime.datetime(2020, 1, 1, 18, 0, 0), once_timer._next_execution())
            once_timer = timer.At(hour=[13, 17, 20])
            self._assert_datetime(datetime.datetime(2020, 1, 1, 17, 0, 0), once_timer._next_execution())
            once_timer = timer.At(hour=None, minute=18)
            self._assert_datetime(datetime.datetime(2020, 1, 1, 15, 18, 0), once_timer._next_execution())
            once_timer = timer.At(hour=timer.EveryNth(3), minute=None, second=None)
            self._assert_datetime(datetime.datetime(2020, 1, 1, 15, 7, 17), once_timer._next_execution())

    def test_stepback(self) -> None:
        with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
            once_timer = timer.At(hour=None, minute=0)
            self._assert_datetime(datetime.datetime(2020, 1, 1, 16, 0, 0), once_timer._next_execution())
            once_timer = timer.At(hour=15, minute=0)
            self._assert_datetime(datetime.datetime(2020, 1, 2, 15, 0, 0), once_timer._next_execution())
            once_timer = timer.At(month=1, day=1, hour=timer.EveryNth(15), minute=[5, 7], second=16)
            self._assert_datetime(datetime.datetime(2021, 1, 1, 0, 5, 16), once_timer._next_execution())

    def test_overflows(self) -> None:
        with ClockMock(datetime.datetime(2020, 12, 31, 23, 59, 46)) as clock:
            once_timer = timer.At(hour=None, minute=None, second=timer.EveryNth(15))
            self._assert_datetime(datetime.datetime(2021, 1, 1, 0, 0, 0), once_timer._next_execution())
        with ClockMock(datetime.datetime(2019, 2, 1, 15, 7, 17)) as clock:
            once_timer = timer.At(day=29)
            self._assert_datetime(datetime.datetime(2019, 3, 29, 0, 0, 0), once_timer._next_execution())
        with ClockMock(datetime.datetime(2020, 2, 1, 15, 7, 17)) as clock:
            once_timer = timer.At(day=29)
            self._assert_datetime(datetime.datetime(2020, 2, 29, 0, 0, 0), once_timer._next_execution())
        with ClockMock(datetime.datetime(2020, 4, 1, 15, 7, 17)) as clock:
            once_timer = timer.At(day=31)
            self._assert_datetime(datetime.datetime(2020, 5, 31, 0, 0, 0), once_timer._next_execution())
        with ClockMock(datetime.datetime(2019, 1, 1, 15, 7, 17)) as clock:
            once_timer = timer.At(weeknum=53)
            self._assert_datetime(datetime.datetime(2020, 12, 28, 0, 0, 0), once_timer._next_execution())
        with ClockMock(datetime.datetime(2020, 1, 1, 0, 0, 0)) as clock:
            once_timer = timer.At(year=2019)
            self.assertIsNone(once_timer._next_execution())

    def test_exception(self) -> None:
        with self.assertRaises(ValueError):
            once_timer = timer.At(day=[1, 5, 15], weeknum=timer.EveryNth(2))


class BoolTimerTest(unittest.TestCase):
    @async_test
    async def test_ton(self) -> None:
        begin = datetime.datetime(2020, 12, 31, 23, 59, 46)
        call_times = []

        def save_time(*args):
            call_times.append(datetime.datetime.now())

        base = ExampleSubscribable(bool)
        ton = timer.TOn(base, datetime.timedelta(seconds=42))

        with unittest.mock.patch.object(ton, "_publish", new=Mock(side_effect=save_time)) as publish_mock:
            with ClockMock(begin, actual_sleep=0.05) as clock:
                self.assertFalse(await ton.read())

                # False should not be forwarded, when value is already False
                await base.publish(False, [self])
                await asyncio.sleep(0.01)
                publish_mock.assert_not_called()

                # True should be forwarded with delay
                await base.publish(True, [self])
                await asyncio.sleep(45)
                publish_mock.assert_called_with(True, unittest.mock.ANY)
                self.assertAlmostEqual(begin + datetime.timedelta(seconds=42.01), call_times[-1],
                                       delta=datetime.timedelta(seconds=.01))

                # False should now be forwarded immediately
                publish_mock.reset_mock()
                await base.publish(False, [self])
                await asyncio.sleep(0.01)
                publish_mock.assert_called_with(False, unittest.mock.ANY)

                # True delay should be suppressable with False ...
                publish_mock.reset_mock()
                await base.publish(True, [self])
                await asyncio.sleep(20)
                await base.publish(False, [self])
                await asyncio.sleep(30)
                publish_mock.assert_not_called()

                # ... and not extendable with True
                publish_mock.reset_mock()
                tic = datetime.datetime.now()
                await base.publish(True, [self])
                await asyncio.sleep(20)
                await base.publish(True, [self])
                await asyncio.sleep(30)
                publish_mock.assert_called_with(True, unittest.mock.ANY)
                self.assertAlmostEqual(tic + datetime.timedelta(seconds=42), call_times[-1],
                                       delta=datetime.timedelta(seconds=.01))

    @async_test
    async def test_toff(self) -> None:
        begin = datetime.datetime(2020, 12, 31, 23, 59, 46)
        call_times = []

        def save_time(*args):
            call_times.append(datetime.datetime.now())

        base = ExampleSubscribable(bool)
        toff = timer.TOff(base, datetime.timedelta(seconds=42))

        with unittest.mock.patch.object(toff, "_publish", new=Mock(side_effect=save_time)) as publish_mock:
            with ClockMock(begin, actual_sleep=0.05) as clock:
                self.assertFalse(await toff.read())

                # False should not be forwarded, when value is already False
                await base.publish(False, [self])
                await asyncio.sleep(0.01)
                publish_mock.assert_not_called()

                # True should be forwarded with immediately
                await base.publish(True, [self])
                await asyncio.sleep(0.01)
                publish_mock.assert_called_with(True, unittest.mock.ANY)

                # False should now be forwarded with delay
                publish_mock.reset_mock()
                await base.publish(False, [self])
                await asyncio.sleep(45)
                publish_mock.assert_called_with(False, unittest.mock.ANY)
                self.assertAlmostEqual(begin + datetime.timedelta(seconds=42.01), call_times[-1],
                                       delta=datetime.timedelta(seconds=.01))

                # False delay should be suppressable with True ...
                await base.publish(True, [self])
                await asyncio.sleep(0.01)
                publish_mock.reset_mock()
                await base.publish(False, [self])
                await asyncio.sleep(20)
                await base.publish(True, [self])
                await asyncio.sleep(30)
                publish_mock.assert_not_called()

                # ... and not extendable with False
                publish_mock.reset_mock()
                tic = datetime.datetime.now()
                await base.publish(False, [self])
                await asyncio.sleep(20)
                await base.publish(False, [self])
                await asyncio.sleep(30)
                publish_mock.assert_called_with(False, unittest.mock.ANY)
                self.assertAlmostEqual(tic + datetime.timedelta(seconds=42), call_times[-1],
                                       delta=datetime.timedelta(seconds=.01))

    @async_test
    async def test_pulse(self) -> None:
        begin = datetime.datetime(2020, 12, 31, 23, 59, 46)
        call_times = []

        def save_time(*args):
            call_times.append(datetime.datetime.now())

        base = ExampleSubscribable(bool)
        tpulse = timer.TPulse(base, datetime.timedelta(seconds=42))

        with unittest.mock.patch.object(tpulse, "_publish", new=Mock(side_effect=save_time)) as publish_mock:
            with ClockMock(begin, actual_sleep=0.05) as clock:
                self.assertFalse(await tpulse.read())

                # False should not be forwarded
                await base.publish(False, [self])
                await asyncio.sleep(0.01)
                publish_mock.assert_not_called()

                # True should be forwarded with immediately and automatically result in a later False
                await base.publish(True, [self])
                await asyncio.sleep(0.01)
                publish_mock.assert_called_with(True, unittest.mock.ANY)
                await asyncio.sleep(45)
                self.assertEqual(2, publish_mock.call_count)
                publish_mock.assert_called_with(False, unittest.mock.ANY)
                self.assertAlmostEqual(begin + datetime.timedelta(seconds=42.01), call_times[-1],
                                       delta=datetime.timedelta(seconds=.01))

                # Pulse should not be stoppable with False or extendable with a second rising edge
                publish_mock.reset_mock()
                await base.publish(False, [self])
                await asyncio.sleep(0.01)
                publish_mock.assert_not_called()
                start = clock.now()
                await base.publish(True, [self])
                await asyncio.sleep(0.01)
                publish_mock.assert_called_with(True, unittest.mock.ANY)
                # second pulse after 20s
                await asyncio.sleep(20)
                await base.publish(False, [self])
                await asyncio.sleep(0.01)
                self.assertEqual(1, publish_mock.call_count)
                await asyncio.sleep(20)
                await base.publish(True, [self])
                await asyncio.sleep(0.01)
                self.assertEqual(1, publish_mock.call_count)
                # Wait 25s more
                await asyncio.sleep(25)
                self.assertEqual(2, publish_mock.call_count)
                publish_mock.assert_called_with(False, unittest.mock.ANY)
                self.assertAlmostEqual(start + datetime.timedelta(seconds=42.01), call_times[-1],
                                       delta=datetime.timedelta(seconds=.01))

                # And a two consecutive True values should not trigger a second pulse
                publish_mock.reset_mock()
                await base.publish(True, [self])
                await asyncio.sleep(0.2)
                publish_mock.assert_not_called()

                # ... but a False -> True edge should
                await base.publish(False, [self])
                await asyncio.sleep(0.2)
                publish_mock.assert_not_called()
                start = clock.now()
                await base.publish(True, [self])
                await asyncio.sleep(0.01)
                self.assertEqual(1, publish_mock.call_count)
                await asyncio.sleep(45)
                self.assertEqual(2, publish_mock.call_count)
                publish_mock.assert_called_with(False, unittest.mock.ANY)
                self.assertAlmostEqual(start + datetime.timedelta(seconds=42.01), call_times[-1],
                                       delta=datetime.timedelta(seconds=.01))


class TimerSwitchTest(unittest.TestCase):
    @async_test
    async def test_simple(self) -> None:
        pub_on1 = ExampleSubscribable(type(None))
        pub_on2 = ExampleSubscribable(type(None))
        pub_off1 = ExampleSubscribable(type(None))
        pub_off2 = ExampleSubscribable(type(None))

        timerswitch = timer.TimerSwitch([pub_on1, pub_on2], [pub_off1, pub_off2])

        with unittest.mock.patch.object(timerswitch, "_publish") as publish_mock:
            self.assertFalse(await timerswitch.read())

            await pub_on1.publish(None, [self])
            publish_mock.assert_called_once_with(True, unittest.mock.ANY)
            self.assertTrue(await timerswitch.read())

            publish_mock.reset_mock()
            await pub_on2.publish(None, [self])
            publish_mock.assert_not_called()
            self.assertTrue(await timerswitch.read())

            await pub_off2.publish(None, [self])
            publish_mock.assert_called_once_with(False, unittest.mock.ANY)
            self.assertFalse(await timerswitch.read())

            publish_mock.reset_mock()
            await pub_off1.publish(None, [self])
            publish_mock.assert_not_called()
            self.assertFalse(await timerswitch.read())

    @async_test
    async def test_duration(self) -> None:
        begin = datetime.datetime(2020, 1, 1, 0, 0, 0)
        pub_on1 = ExampleSubscribable(type(None))
        pub_on2 = ExampleSubscribable(type(None))

        timerswitch = timer.TimerSwitch([pub_on1, pub_on2], duration=datetime.timedelta(seconds=42))

        with unittest.mock.patch.object(timerswitch, "_publish") as publish_mock:
            with ClockMock(begin, actual_sleep=0.05) as clock:
                self.assertFalse(await timerswitch.read())

                await pub_on2.publish(None, [self])
                publish_mock.assert_called_once_with(True, unittest.mock.ANY)
                self.assertTrue(await timerswitch.read())

                # Retrigger duration after 32 seconds
                await asyncio.sleep(32)
                publish_mock.reset_mock()
                await pub_on1.publish(None, [self])
                publish_mock.assert_not_called()
                self.assertTrue(await timerswitch.read())

                # After 52 seconds, the timerswitch should still be on, since it has been retriggered at 32 sec.
                await asyncio.sleep(20)
                publish_mock.assert_not_called()
                self.assertTrue(await timerswitch.read())

                # After 74 seconds the timerswitch should automatically switch off. Let't test its state at 73 and 75
                # sec.
                await asyncio.sleep(21)
                publish_mock.assert_not_called()
                self.assertTrue(await timerswitch.read())

                await asyncio.sleep(2)
                publish_mock.assert_called_once_with(False, unittest.mock.ANY)
                self.assertFalse(await timerswitch.read())

    def test_error(self) -> None:
        pub_on1 = ExampleSubscribable(bool)
        pub_off1 = ExampleSubscribable(bool)

        with self.assertRaises(ValueError):
            timer.TimerSwitch([pub_on1], [pub_off1], datetime.timedelta(seconds=42))


class RateLimitedSubscriptionTest(unittest.TestCase):
    @async_test
    async def test_simple(self) -> None:
        begin = datetime.datetime(2020, 12, 31, 23, 59, 46)
        call_times = []

        def save_time(*args):
            call_times.append(datetime.datetime.now())

        base = ExampleSubscribable(int)
        rate_limiter = timer.RateLimitedSubscription(base, 1.0)

        self.assertIs(rate_limiter.type, int)

        with unittest.mock.patch.object(rate_limiter, "_publish", new=Mock(side_effect=save_time)) as publish_mock:
            with ClockMock(begin, actual_sleep=0.05) as clock:
                # First value should be forwarded immediately
                await base.publish(42, [self])
                await asyncio.sleep(0.1)
                publish_mock.assert_called_once_with(42, unittest.mock.ANY)
                publish_mock.reset_mock()
                await asyncio.sleep(0.4)

                # Second value within min_interval should be hold back for remaining interval time (1-0.1-0.4 = 0.5)
                # and be superseded by third value
                await base.publish(56, [self])
                await asyncio.sleep(0.1)
                publish_mock.assert_not_called()

                await base.publish(21, [unittest.mock.sentinel, self])
                await asyncio.sleep(0.5)
                publish_mock.assert_called_once_with(21, unittest.mock.ANY)
                self.assertEqual(21, publish_mock.call_args[0][0])
                self.assertIs(publish_mock.call_args[0][1][0], unittest.mock.sentinel)

                # After longer period of time there should not be new calls
                await asyncio.sleep(5)
                publish_mock.assert_called_once()

                # A new value should now be forwarded immediately, again
                publish_mock.reset_mock()
                await base.publish(42, [self])
                await asyncio.sleep(0.1)
                publish_mock.assert_called_once_with(42, unittest.mock.ANY)


class RampTest(unittest.TestCase):
    @async_test
    async def test_simple(self) -> None:
        begin = datetime.datetime(2020, 12, 31, 23, 59, 46)
        BLACK = datatypes.RGBWUInt8(
            datatypes.RGBUInt8(datatypes.RangeUInt8(0), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0)),
            datatypes.RangeUInt8(0))
        RED = datatypes.RGBWUInt8(
            datatypes.RGBUInt8(datatypes.RangeUInt8(255), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0)),
            datatypes.RangeUInt8(0))
        MED_WHITE = datatypes.RGBWUInt8(
            datatypes.RGBUInt8(datatypes.RangeUInt8(0), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0)),
            datatypes.RangeUInt8(127))

        subscribable1 = ExampleSubscribable(datatypes.RangeUInt8)
        ramp1 = timer.IntRamp(subscribable1, datetime.timedelta(seconds=1), max_frequency=2)
        writable1 = ExampleWritable(datatypes.RangeUInt8).connect(ramp1)
        subscribable2 = ExampleSubscribable(datatypes.RangeFloat1)
        ramp2 = timer.FloatRamp(subscribable2, datetime.timedelta(seconds=1), max_frequency=4)
        writable2 = ExampleWritable(datatypes.RangeFloat1).connect(ramp2)
        subscribable3 = ExampleSubscribable(datatypes.RGBWUInt8)
        ramp3 = timer.RGBWHSVRamp(subscribable3, datetime.timedelta(seconds=2), max_frequency=2)
        writable3 = ExampleWritable(datatypes.RGBWUInt8).connect(ramp3)

        self.assertIs(ramp1.type, datatypes.RangeUInt8)
        with self.assertRaises(shc.base.UninitializedError):
            await ramp2.read()

        with ClockMock(begin, actual_sleep=0.05) as clock:
            await subscribable1.publish(datatypes.RangeUInt8(0), [self])
            await subscribable2.publish(datatypes.RangeFloat1(0), [self])
            await subscribable3.publish(BLACK, [self])
            await asyncio.sleep(0.05)
            writable1._write.assert_called_once_with(datatypes.RangeUInt8(0), [self, subscribable1, ramp1])
            writable2._write.assert_called_once_with(datatypes.RangeFloat1(0), [self, subscribable2, ramp2])
            writable3._write.assert_called_once_with(BLACK, [self, subscribable3, ramp3])
            self.assertEqual(BLACK, await ramp3.read())

            writable1._write.reset_mock()
            writable2._write.reset_mock()
            writable3._write.reset_mock()

            await subscribable1.publish(datatypes.RangeUInt8(255), [self])
            await subscribable2.publish(datatypes.RangeFloat1(1), [self])
            await subscribable3.publish(RED, [self])
            await asyncio.sleep(0.05)
            # Assert first step
            writable1._write.assert_called_once_with(datatypes.RangeUInt8(128), [ramp1])
            writable2._write.assert_called_once_with(datatypes.RangeFloat1(0.25), [ramp2])
            writable3._write.assert_called_once_with(datatypes.RGBWUInt8(
                datatypes.RGBUInt8(datatypes.RangeUInt8(64), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0)),
                datatypes.RangeUInt8(0)),
                [ramp3])

            # Wait for and assert second step
            await asyncio.sleep(0.5)
            writable1._write.assert_called_with(datatypes.RangeUInt8(255), [ramp1])
            writable2._write.assert_called_with(datatypes.RangeFloat1(0.75), [ramp2])
            writable3._write.assert_called_with(datatypes.RGBWUInt8(
                datatypes.RGBUInt8(datatypes.RangeUInt8(128), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0)),
                datatypes.RangeUInt8(0)),
                [ramp3])
            self.assertEqual(datatypes.RangeUInt8(255), await ramp1.read())
            self.assertEqual(datatypes.RGBWUInt8(
                datatypes.RGBUInt8(datatypes.RangeUInt8(128), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0)),
                datatypes.RangeUInt8(0)),
                await ramp3.read())

            # Let's interrupt ramp3 and ramp to MED_WHITE
            # (both channels only ramp half the way, so it should reduce the duration to 1s
            writable3._write.reset_mock()
            await subscribable3.publish(MED_WHITE, [self])
            await asyncio.sleep(0.5)
            writable3._write.assert_called_once_with(datatypes.RGBWUInt8(
                datatypes.RGBUInt8(datatypes.RangeUInt8(85), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0)),
                datatypes.RangeUInt8(42)),
                [ramp3])

            # (due to rounding errors and precautions, it uses 3 instead of 2 steps for this ramp)
            await asyncio.sleep(0.66)
            writable3._write.assert_called_with(MED_WHITE, [ramp3])

            # Assert end result
            await asyncio.sleep(5)
            writable1._write.assert_called_with(datatypes.RangeUInt8(255), [ramp1])
            writable2._write.assert_called_with(datatypes.RangeFloat1(1.0), [ramp2])
            writable3._write.assert_called_with(MED_WHITE, [ramp3])

    @async_test
    async def test_enable_ramp(self) -> None:
        begin = datetime.datetime(2020, 12, 31, 23, 59, 46)
        BLACK = datatypes.RGBUInt8(datatypes.RangeUInt8(0), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0))
        RED = datatypes.RGBUInt8(datatypes.RangeUInt8(255), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0))

        subscribable1 = ExampleSubscribable(datatypes.RGBUInt8)
        enable1 = ExampleReadable(bool, True)
        ramp1 = timer.RGBHSVRamp(subscribable1, datetime.timedelta(seconds=1), max_frequency=2, enable_ramp=enable1)
        writable1 = ExampleWritable(datatypes.RGBUInt8).connect(ramp1)

        with ClockMock(begin, actual_sleep=0.05) as clock:
            # Ramping should work normal (as in the other test
            await subscribable1.publish(BLACK, [self])
            writable1._write.assert_called_once_with(BLACK, [self, subscribable1, ramp1])
            self.assertEqual(BLACK, await ramp1.read())

            writable1._write.reset_mock()

            await subscribable1.publish(RED, [self])
            await asyncio.sleep(0.05)
            writable1._write.assert_called_once_with(
                datatypes.RGBUInt8(datatypes.RangeUInt8(128), datatypes.RangeUInt8(0), datatypes.RangeUInt8(0)),
                [ramp1])
            await asyncio.sleep(0.5)
            writable1._write.assert_called_with(RED, [ramp1])
            self.assertEqual(RED, await ramp1.read())

            writable1._write.reset_mock()
            await asyncio.sleep(0.5)
            writable1._write.assert_not_called()

            # ... until we disable the ramping
            enable1.read.return_value = False

            await subscribable1.publish(BLACK, [self])
            await asyncio.sleep(0.05)
            writable1._write.assert_called_once_with(BLACK, [self, subscribable1, ramp1])
            self.assertEqual(BLACK, await ramp1.read())

            writable1._write.reset_mock()
            await asyncio.sleep(0.5)
            writable1._write.assert_not_called()

    @async_test
    async def test_stateful_target(self) -> None:
        begin = datetime.datetime(2020, 12, 31, 23, 59, 46)

        subscribable1 = ExampleSubscribable(datatypes.RangeUInt8)
        ramp1 = timer.IntRamp(subscribable1, datetime.timedelta(seconds=1), max_frequency=2, dynamic_duration=False)
        variable1 = shc.Variable(datatypes.RangeUInt8).connect(ramp1)
        writable1 = ExampleWritable(datatypes.RangeUInt8).connect(variable1)

        with ClockMock(begin, actual_sleep=0.05) as clock:
            await subscribable1.publish(datatypes.RangeUInt8(0), [self])
            await asyncio.sleep(0.05)
            writable1._write.assert_called_once_with(datatypes.RangeUInt8(0), [self, subscribable1, ramp1, variable1])
            writable1._write.reset_mock()

            await subscribable1.publish(datatypes.RangeUInt8(255), [self])
            await asyncio.sleep(0.05)
            # Assert first step
            writable1._write.assert_called_once_with(datatypes.RangeUInt8(128), [ramp1, variable1])
            writable1._write.reset_mock()

            # Let's interrupt the ramp be sending a new value directly to the Variable
            await asyncio.sleep(0.25)
            await variable1.write(datatypes.RangeUInt8(192), [self])

            await asyncio.sleep(1.0)
            writable1._write.assert_called_once_with(datatypes.RangeUInt8(192), [self, variable1])
            writable1._write.reset_mock()

            # And now, let's do a new ramp to 0, which should start at 192
            await subscribable1.publish(datatypes.RangeUInt8(0), [self])
            await asyncio.sleep(0.05)
            writable1._write.assert_called_once_with(datatypes.RangeUInt8(96), [ramp1, variable1])
            await asyncio.sleep(0.5)
            writable1._write.assert_called_with(datatypes.RangeUInt8(0), [ramp1, variable1])

    @async_test
    async def test_fade_step_ramp(self) -> None:
        begin = datetime.datetime(2020, 12, 31, 23, 59, 46)

        subscribable1 = ExampleSubscribable(datatypes.FadeStep)
        ramp1 = timer.FadeStepRamp(subscribable1, datetime.timedelta(seconds=1), max_frequency=2,
                                   dynamic_duration=False)
        variable1 = shc.Variable(datatypes.RangeFloat1).connect(ramp1)
        writable1 = ExampleWritable(datatypes.RangeFloat1).connect(variable1)

        with ClockMock(begin, actual_sleep=0.05) as clock:
            with self.assertLogs() as logs:
                await subscribable1.publish(datatypes.FadeStep(0.5), [self])
                await asyncio.sleep(0.05)
            self.assertIn("Cannot apply FadeStep", logs.records[0].msg)  # type: ignore
            writable1._write.assert_not_called()
            writable1._write.reset_mock()

            await variable1.write(datatypes.RangeFloat1(0.0), [self])
            await asyncio.sleep(0.05)
            writable1._write.reset_mock()

            await subscribable1.publish(datatypes.FadeStep(0.5), [self])
            await asyncio.sleep(0.05)
            # Assert first step
            writable1._write.assert_called_once_with(datatypes.RangeFloat1(0.25), [ramp1, variable1])
            writable1._write.reset_mock()

            # Let's interrupt the ramp be sending a new value directly to the varible
            await asyncio.sleep(0.25)
            await variable1.write(datatypes.RangeFloat1(0.75), [self])

            await asyncio.sleep(1.0)
            writable1._write.assert_called_once_with(datatypes.RangeFloat1(0.75), [self, variable1])
            writable1._write.reset_mock()

            # And now, let's do a new ramp to 0.25, which should start at 0.75
            await subscribable1.publish(datatypes.FadeStep(-0.5), [self])
            await asyncio.sleep(0.05)
            writable1._write.assert_called_once_with(datatypes.RangeFloat1(0.5), [ramp1, variable1])
            await asyncio.sleep(0.5)
            writable1._write.assert_called_with(datatypes.RangeFloat1(0.25), [ramp1, variable1])



#!/usr/bin/env python
# Copyright 2020-2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import subprocess
from setuptools import setup
from setuptools.command.build_py import build_py


class custom_build_py(build_py):
    def run(self):
        subprocess.check_call(['npm', 'install'])
        subprocess.check_call(['npm', 'run', 'build'])
        super().run()


setup(
    cmdclass={
        'build_py': custom_build_py
    },
)



import asyncio
import unittest
import unittest.mock
import warnings
from pathlib import Path
from typing import NamedTuple

import mypy.api

from shc import variables, base, expressions
from ._helper import async_test, ExampleReadable, ExampleWritable, AsyncMock


class SimpleVariableTest(unittest.TestCase):

    @unittest.mock.patch('shc.variables.Variable._publish')
    @async_test
    async def test_basic_behaviour(self, publish_patch):
        var = variables.Variable(int)
        await var.write(42, [self])
        publish_patch.assert_called_once_with(42, [self])
        self.assertEqual(42, await var.read())

        publish_patch.reset_mock()
        await var.write(21, [self])
        publish_patch.assert_called_once_with(21, [self])
        self.assertEqual(21, await var.read())

        publish_patch.reset_mock()
        await var.write(21, [self])
        publish_patch.assert_not_called()

    @async_test
    async def test_static_initialization(self):
        var = variables.Variable(int)
        with self.assertRaises(base.UninitializedError):
            await var.read()

        var = variables.Variable(int, initial_value=42)
        self.assertEqual(42, await var.read())
        await var.write(21, [self])
        self.assertEqual(21, await var.read())

    @async_test
    async def test_read_initialization(self):
        var1 = variables.Variable(int)
        var2 = variables.Variable(int, initial_value=21)
        var3 = variables.Variable(int)
        source = ExampleReadable(int, 42)
        var1.set_provider(source)

        with self.assertRaises(base.UninitializedError):
            await var1.read()
        self.assertEqual(21, await var2.read())
        with self.assertRaises(base.UninitializedError):
            await var3.read()

        await variables.read_initialize_variables()
        self.assertEqual(42, await var1.read())
        self.assertEqual(21, await var2.read())
        with self.assertRaises(base.UninitializedError):
            await var3.read()

    @async_test
    async def test_expression_wrapper(self):
        var = variables.Variable(int, initial_value=42)
        wrapper = var.EX
        subscriber = ExampleWritable(int)
        wrapper.subscribe(subscriber)

        self.assertIsInstance(wrapper, expressions.ExpressionBuilder)
        self.assertEqual(42, await wrapper.read())
        await var.write(21, [self])
        await asyncio.sleep(0.01)
        subscriber._write.assert_called_with(21, [self, var])


class ExampleTupleType(NamedTuple):
    a: int
    b: float


class ExampleRecursiveTupleType(NamedTuple):
    a: ExampleTupleType
    b: int


class VariableFieldsTest(unittest.TestCase):
    @async_test
    async def test_field_subscribing(self):
        var = variables.Variable(ExampleTupleType)
        field_subscriber = ExampleWritable(int)
        var.field('a').subscribe(field_subscriber)

        with self.assertRaises(TypeError):
            var2 = variables.Variable(float)
            await var2.field('c')  # type: ignore

        with self.assertRaises(KeyError):
            await var.field('c')  # type: ignore

        with self.assertRaises(base.UninitializedError):
            await var.field('a').read()

        await var.write(ExampleTupleType(42, 3.1416), [self])
        self.assertEqual(42, await var.field('a').read())
        await asyncio.sleep(0.01)
        field_subscriber._write.assert_called_once_with(42, [self, var.field('a')])

        field_subscriber._write.reset_mock()
        await var.write(ExampleTupleType(42, 2.719), [self])
        field_subscriber._write.assert_not_called()

    @async_test
    async def test_field_writing(self):
        var = variables.Variable(ExampleTupleType, name="A test variable")
        field_subscriber = ExampleWritable(int)
        other_field_subscriber = ExampleWritable(float)
        subscriber = ExampleWritable(ExampleTupleType)
        var.subscribe(subscriber)
        var.field('a').subscribe(field_subscriber)
        var.field('b').subscribe(other_field_subscriber)

        with self.assertRaises(base.UninitializedError):
            await var.field('a').write(21, [self])

        await var.write(ExampleTupleType(42, 3.1416), [self])
        await asyncio.sleep(0.01)

        subscriber._write.reset_mock()
        field_subscriber._write.reset_mock()
        other_field_subscriber._write.reset_mock()

        await var.field('a').write(21, [self])
        self.assertEqual(21, await var.field('a').read())
        await asyncio.sleep(0.01)
        subscriber._write.assert_called_once_with(ExampleTupleType(21, 3.1416), [self, var.field('a'), var])
        field_subscriber._write.assert_called_once_with(21, [self, var.field('a'), var.field('a')])
        other_field_subscriber._write.assert_not_called()

    @async_test
    async def test_recursive_field_subscribing(self):
        var = variables.Variable(ExampleRecursiveTupleType)
        field_subscriber = ExampleWritable(int)
        var.field('a').field('a').subscribe(field_subscriber)

        with self.assertRaises(KeyError):
            await var.field('a').field('c')  # type: ignore

        with self.assertRaises(base.UninitializedError):
            await var.field('a').field('a').read()

        await var.write(ExampleRecursiveTupleType(ExampleTupleType(42, 3.1416), 7), [self])
        self.assertEqual(42, await var.field('a').field('a').read())
        await asyncio.sleep(0.01)
        field_subscriber._write.assert_called_once_with(42, [self, var.field('a').field('a')])

        field_subscriber._write.reset_mock()
        await var.write(ExampleRecursiveTupleType(ExampleTupleType(42, 3.1416), 6), [self])
        await var.write(ExampleRecursiveTupleType(ExampleTupleType(42, 2.719), 6), [self])
        field_subscriber._write.assert_not_called()

    @async_test
    async def test_recursive_field_writing_legacy(self):
        var = variables.Variable(ExampleRecursiveTupleType)
        subscriber = ExampleWritable(ExampleRecursiveTupleType)
        intermediate_subscriber = ExampleWritable(ExampleTupleType)
        field_subscriber = ExampleWritable(int)
        other_field_subscriber = ExampleWritable(float)
        another_field_subscriber = ExampleWritable(int)

        with warnings.catch_warnings(record=True) as w:
            var.subscribe(subscriber)
            var.a.subscribe(intermediate_subscriber)
            var.a.a.subscribe(field_subscriber)
            var.a.b.subscribe(other_field_subscriber)
            var.b.subscribe(another_field_subscriber)
        self.assertGreaterEqual(len(w), 4)
        self.assertTrue(issubclass(w[0].category, DeprecationWarning))

        with warnings.catch_warnings(record=True):  # There seems to be a bug, that record=False breaks the catching
            with self.assertRaises(AttributeError):
                var_c = var.c
            with self.assertRaises(AttributeError):
                var_c = var.a.c

            with self.assertRaises(base.UninitializedError):
                await var.a.a.write(21, [self])

            await var.write(ExampleRecursiveTupleType(ExampleTupleType(42, 3.1416), 7), [self])
            await asyncio.sleep(0.01)

            subscriber._write.reset_mock()
            intermediate_subscriber._write.reset_mock()
            field_subscriber._write.reset_mock()
            other_field_subscriber._write.reset_mock()
            another_field_subscriber._write.reset_mock()

            await var.a.a.write(21, [self])
            self.assertEqual(21, await var.a.a.read())
            self.assertEqual(ExampleTupleType(21, 3.1416), await var.a.read())
            await asyncio.sleep(0.01)
            subscriber._write.assert_called_once_with(ExampleRecursiveTupleType(ExampleTupleType(21, 3.1416), 7),
                                                      [self, var.a.a, var.a, var])
            intermediate_subscriber._write.assert_called_once_with(
                ExampleTupleType(21, 3.1416), [self, var.a.a, var.a, var.a])
            field_subscriber._write.assert_called_once_with(21, [self, var.a.a, var.a, var.a.a])
            other_field_subscriber._write.assert_not_called()
            another_field_subscriber._write.assert_not_called()

    @async_test
    async def test_recursive_field_writing(self):
        var = variables.Variable(ExampleRecursiveTupleType)
        subscriber = ExampleWritable(ExampleRecursiveTupleType)
        intermediate_subscriber = ExampleWritable(ExampleTupleType)
        field_subscriber = ExampleWritable(int)
        other_field_subscriber = ExampleWritable(float)
        another_field_subscriber = ExampleWritable(int)
        var.subscribe(subscriber)
        var.field('a').subscribe(intermediate_subscriber)
        var.field('a').field('a').subscribe(field_subscriber)
        var.field('a').field('b').subscribe(other_field_subscriber)
        var.field('b').subscribe(another_field_subscriber)

        with self.assertRaises(base.UninitializedError):
            await var.field('a').field('a').write(21, [self])

        await var.write(ExampleRecursiveTupleType(ExampleTupleType(42, 3.1416), 7), [self])
        await asyncio.sleep(0.01)

        subscriber._write.reset_mock()
        intermediate_subscriber._write.reset_mock()
        field_subscriber._write.reset_mock()
        other_field_subscriber._write.reset_mock()
        another_field_subscriber._write.reset_mock()

        await var.field('a').field('a').write(21, [self])
        self.assertEqual(21, await var.field('a').field('a').read())
        self.assertEqual(ExampleTupleType(21, 3.1416), await var.field('a').read())
        await asyncio.sleep(0.01)
        subscriber._write.assert_called_once_with(ExampleRecursiveTupleType(ExampleTupleType(21, 3.1416), 7),
                                                  [self, var.field('a').field('a'), var.field('a'), var])
        intermediate_subscriber._write.assert_called_once_with(
            ExampleTupleType(21, 3.1416),
            [self, var.field('a').field('a'), var.field('a'), var.field('a')])
        field_subscriber._write.assert_called_once_with(
            21,
            [self, var.field('a').field('a'), var.field('a'), var.field('a').field('a')])
        other_field_subscriber._write.assert_not_called()
        another_field_subscriber._write.assert_not_called()

    @async_test
    async def test_expression_wrapper(self):
        var = variables.Variable(ExampleTupleType, initial_value=ExampleTupleType(42, 3.1416))
        wrapper = var.field('a').EX
        subscriber = ExampleWritable(int)
        wrapper.subscribe(subscriber)

        self.assertIsInstance(wrapper, expressions.ExpressionBuilder)
        self.assertEqual(42, await wrapper.read())
        await var.field('a').write(21, [self])
        await asyncio.sleep(0.01)
        subscriber._write.assert_called_with(21, [self, var.field('a'), var.field('a')])


class ConnectedVariablesTest(unittest.TestCase):
    @async_test
    async def test_simple_concurrent_update(self) -> None:
        var1 = variables.Variable(int)
        var2 = variables.Variable(int).connect(var1)

        await asyncio.gather(var1.write(42, []), var2.write(56, []))
        await asyncio.sleep(0.01)
        self.assertEqual(await var1.read(), await var2.read())

    @async_test
    async def test_concurrent_field_update_publishing(self) -> None:
        var1 = variables.Variable(ExampleTupleType)
        var2 = variables.Variable(ExampleTupleType).connect(var1)
        var3 = variables.Variable(int).connect(var2.field('a'))

        writable1 = ExampleWritable(int).connect(var1.field('a'))  # TODO add different delays
        writable3 = ExampleWritable(int).connect(var3)

        await asyncio.gather(var1.write(ExampleTupleType(42, 3.1416), []), var3.write(56, []))
        await asyncio.sleep(0.01)
        self.assertEqual(await var1.field('a').read(), await var3.read())

        self.assertLessEqual(writable1._write.call_count, 3)
        self.assertLessEqual(writable3._write.call_count, 3)
        # 1st arg of 2nd call shall be equal
        self.assertEqual(writable1._write.call_args[0][0], writable3._write.call_args[0][0])


class MyPyPluginTest(unittest.TestCase):
    def test_mypy_plugin(self) -> None:
        asset_dir = Path(__file__).parent / 'assets' / 'mypy_plugin_test'
        result = mypy.api.run(['--config-file',
                               str(asset_dir / 'mypy.ini'),
                               str(asset_dir / 'variable_typing_example.py')])
        self.assertIn("variable_typing_example.py:22: error", result[0])
        self.assertNotIn("variable_typing_example.py:23: error", result[0])
        self.assertIn("variable_typing_example.py:25: error", result[0])
        self.assertNotIn("variable_typing_example.py:26: error", result[0])
        self.assertIn("variable_typing_example.py:27: error", result[0])
        self.assertIn("variable_typing_example.py:28: error", result[0])
        self.assertIn("variable_typing_example.py:29: error", result[0])
        self.assertEqual(1, result[2])



import abc
import asyncio
import logging
import re
from typing import Generic, TypeVar, Set, Type, Optional, List, Pattern, Tuple, Dict, Any, Callable

import aiogram
from aiogram.bot.api import TelegramAPIServer, TELEGRAM_PRODUCTION

from ..base import Writable, Subscribable, Reading, T
from ..supervisor import AbstractInterface

RoleT = TypeVar('RoleT')
UserT = TypeVar('UserT')

logger = logging.getLogger(__name__)


class TelegramBot(AbstractInterface, Generic[UserT, RoleT]):
    """
    The Telegram Bot interface.

    An instance of this class represents a single Telegram chat bot.

    :param api_token: The API token for connecting to the Telegram Bot API. It can be retrieved from Telegram's
        "BotFather" after registering the bot.
    :param auth_provider: An authentication provider to authenticate users by their Telegram user id, check
        authorization and find the chat id of a given user to send them a message.

        When creating connector objects, user role objects that are compatible with the the `auth_provider` must be used
        for authorization. For the :class:`SimpleTelegramAuth` user roles are simple str objects representing the
        authorized users' names, but other `TelegramAuthProvider` implemantations may use more sophisticated user/group
        role objects.
    :param telegram_server: The Telegram Bot API server to connect to. Usually, Telegram's official servers should be
        used (the default value). However, you may also run a local HTTP API, communicating with Telegram's
        network (see https://core.telegram.org/bots/api#using-a-local-bot-api-server). We also use this option for
        testing purposes.
    """

    def __init__(self, api_token: str, auth_provider: "TelegramAuthProvider[UserT, RoleT]",
                 telegram_server: TelegramAPIServer = TELEGRAM_PRODUCTION):
        super().__init__()
        self.auth_provider = auth_provider

        self.bot = aiogram.Bot(token=api_token, server=telegram_server)
        self.dp = aiogram.Dispatcher(self.bot)
        self.dp.register_message_handler(self._handle_start, commands=["start"])
        self.dp.register_message_handler(self._handle_cancel, commands=["cancel"])
        self.dp.register_message_handler(self._handle_select, commands=["s"])
        self.dp.register_callback_query_handler(self._handle_callback_query)
        self.dp.register_message_handler(self._handle_other)

        self.connectors: Dict[str, "TelegramConnector"] = {}
        self.poll_task: Optional[asyncio.Task] = None
        #: Current state/context of each chat. Either None (init state selected) or
        self.chat_state: Dict[int, TelegramConnector] = {}
        #: Maps chat ids to a message id of a messages with an active inline keyboard in that chat, if any.
        #: Used to clean up inline keyboards, when cancelled.
        self.message_with_inline_keyboard: Dict[int, int] = {}

    async def start(self) -> None:
        self.poll_task = asyncio.create_task(self._run())

    async def _run(self) -> None:
        await self.dp.start_polling()

    async def stop(self) -> None:
        self.dp.stop_polling()
        if self.poll_task:
            await self.poll_task
        await self.dp.storage.close()
        await self.dp.storage.wait_closed()
        session = await self.dp.bot.get_session()
        await session.close()

    async def _handle_start(self, message: aiogram.types.Message):
        """
        Handler function for /start command messages
        """
        logger.debug("Received /start command for object from Telegram chat %s", message.chat.id)
        await message.reply("Hi!\nI'm an SHC bot!", reply=False)
        chat_id = message.chat.id
        user = self.auth_provider.get_telegram_user(chat_id)
        if user is None:
            logger.warning("Telegram chat %s is not authorized to do /start", chat_id)
            await message.reply(f"Unauthorized! Please make sure that this Telegram chat's ID ({chat_id}) is "
                                f"authorized to interact with this bot.", reply=False)

    async def _handle_select(self, message: aiogram.types.Message) -> None:
        """
        Handler function for /s command messages for selecting (and reading) a connector
        """
        chat_id = message.chat.id
        logger.debug("Received /s (select) command for object for Telegram chat %s", chat_id)
        user = self.auth_provider.get_telegram_user(chat_id)
        if user is None:
            logger.warning("Telegram chat %s is not authorized to do /s", chat_id)
            await message.reply("Not authorized!")
            return

        # Cancel ongoing value setting (if any)
        if chat_id in self.chat_state:
            await self._do_cancel(chat_id, silent=True)

        connector_id = message.text[3:]  # strip the '/select ' prefix
        if connector_id not in self.connectors:
            logger.debug("Received /s (select) command for non-existing object '%s' for Telegram chat %s", connector_id,
                         message.chat.id)
            await message.reply("Unknown connector/object")
            return
        connector = self.connectors[connector_id]
        not_authorized = True

        # Read value
        if connector.is_readable() and self.auth_provider.has_user_role(user, connector.read_roles):
            not_authorized = False
            read_message = await connector.read_message()
            if read_message:
                await message.reply(read_message, reply=False)

        # Prepare setting value
        if connector.is_settable() and self.auth_provider.has_user_role(user, connector.set_roles):
            not_authorized = False
            # Create custom keyboard/inline keyboard markup
            keyboard = connector.get_setting_keyboard()
            # If no options/custom keyboard is provided, we create an inline keyboard of cancelling the value setting.
            # Otherwise, we add a `/cancel` button to the bottom of the keyboard.
            if keyboard is None:
                keyboard = aiogram.types.InlineKeyboardMarkup(
                    row_width=1,
                    inline_keyboard=[[aiogram.types.InlineKeyboardButton("cancel", callback_data="cancel")]])
                inline_keyboard = True
            else:
                keyboard.keyboard.append([aiogram.types.KeyboardButton("/cancel")])
                inline_keyboard = False

            reply_message = await message.reply(connector.get_set_message(), reply=False, reply_markup=keyboard)
            self.chat_state[message.chat.id] = connector
            if inline_keyboard:
                self.message_with_inline_keyboard[chat_id] = reply_message.message_id

        if not_authorized:
            await message.reply("Not authorized!")

    async def _handle_cancel(self, message: aiogram.types.Message) -> None:
        """
        Handler function for /cancel command messages for cancelling the current chat context
        """
        logger.debug("Received /cancel format for Telegram chat %s", message.chat.id)
        await self._do_cancel(message.chat.id)

    async def _handle_other(self, message: aiogram.types.Message) -> None:
        """
        Handler function for all incoming Telegram messages that are not recognized as a command.

        The message is interpreted as a connector search or a new value for the selected connector, depending on the
        current chat state.
        """
        chat_id = message.chat.id
        user = self.auth_provider.get_telegram_user(chat_id)
        if user is None:
            await message.reply("Not authorized!")
            logger.warning("Received non-command message '%s' from unauthorized Telegram user (chat id %s)",
                           message.text, chat_id)
            return
        logger.debug("Received non-command message '%s' from Telegram user %s", message.text, user)
        if chat_id in self.chat_state:
            await self._do_set(message, user)
        else:
            await self._do_connector_search(message, user)

    async def _do_set(self, message: aiogram.types.Message, user: UserT) -> None:
        """
        Use the given Telegram message as a new value for the currently selected connector in the message's chat

        This method must only be called, if the message's chat is known to have a selected connector/context
        (chat_state) and the chat is known to belong to an authenticated user. The user's authorization for the
        connector is checked by this method.

        :param message: The Telegram message to be handled as a new value
        :param user: The identified user, related to the chat
        """
        chat_id = message.chat.id
        context = self.chat_state.get(chat_id)
        assert(context is not None)  # _do_set() should only be called if a context is present
        if not self.auth_provider.has_user_role(user, context.set_roles):
            logger.warning("User %s is not authorized for setting Telegram connector %s", user, context.name)
            await message.reply("Not authorized!")
            return
        value = message.text
        try:
            logger.info("Received value '%s' for Telegram connector %s", value, context.name)
            context.from_telegram(value)
            await message.reply("🆗", reply=False, reply_markup=aiogram.types.ReplyKeyboardRemove())
            del self.chat_state[chat_id]
        except (TypeError, ValueError) as e:
            logger.warning("Invalid value '%s' received for Telegram connector %s: %s", value, context.name, e)
            await message.reply(f"Invalid value for connector/object {context.name}: {e}")
        except Exception as e:
            logger.error("Error while sending value '%s' to Telegram connector/object %s:", value, context.name,
                         exc_info=e)
            await message.reply(f"Internal error while setting value for connector/object {context.name}")

    async def _do_connector_search(self, message: aiogram.types.Message, user: UserT) -> None:
        """
        Use the given Telegram message as a search term for searching connector objects and ask the user to select
        a connector from the list matching connectors

        This method should only be called, if the message's chat is known to have no a selected connector/context
        (chat_state) yet and the chat is known to belong to an authenticated user.

        :param message: The Telegram message to be handled as a search term
        :param user: The identified user, related to the chat
        """
        connectors = self._find_matching_connectors(message.text, user)
        if connectors:
            await message.reply("Please chose", reply_markup=aiogram.types.ReplyKeyboardMarkup(
                [[aiogram.types.KeyboardButton(f"/s {var.name}")]
                 for var in connectors],
                one_time_keyboard=True,
                resize_keyboard=True), reply=False)
        else:
            await message.reply("No matching connector/object found")

    def _find_matching_connectors(self, search_term: str, user: UserT) -> List["TelegramConnector"]:
        """
        Search for connectors of this bot, matching a given search term, entered by the user, and being accessible for
        the given user.

        This method is used to present a connector selection list to the user, when they enter a search term.
        """
        regex = re.compile(re.escape(search_term), re.IGNORECASE)
        return [var
                for name, var in self.connectors.items()
                if regex.search(name) and (
                    self.auth_provider.has_user_role(user, var.read_roles) and var.is_readable()
                    or self.auth_provider.has_user_role(user, var.set_roles) and var.is_settable())]

    async def _handle_callback_query(self, query: aiogram.types.CallbackQuery) -> None:
        """
        Handler function to be called when an inline button callback query is received from Telegram.

        This is currently only used for "cancel" inline keyboard buttons.
        """
        logger.debug("Received CallbackQuery")
        if query.message is None:
            logger.warning("Received CallbackQuery without message, so the originating chat cannot be identified.")
            return
        if query.data == 'cancel':
            # Remove inline keyboard. Just to be sure. (Should also be done by _do_cancel(), when chat state is tracked
            # correctly)
            chat_id = int(query.message.chat.id)
            message_id = query.message.message_id
            await self.bot.edit_message_reply_markup(chat_id, message_id, reply_markup=None)
            if self.message_with_inline_keyboard.get(chat_id) == message_id:
                del self.message_with_inline_keyboard[chat_id]
            # Now, the actual action cancelling
            await self._do_cancel(chat_id)

    async def _do_cancel(self, chat_id: int, silent: bool = False) -> None:
        """
        Cancel the connector setting in progress in the given chat

        This means:

        * reset the chat_state/unselect the selected context/connector (so following input is not interpreted as a value
          for this connector)
        * remove the custom keyboard (if any)
        * remove the inline keyboard (if any)

        :param chat_id: The Telegram chat id of the chat to be resetted
        :param silent: If True, a message is only sent to the chat if a connector is actually selected in this chat. If
            **there is** a connector selected, a message is sent nonetheless. As a side effect, if no message is sent,
            we cannot reset the custom keyboard either. However, this *should* not be required in this case, if we
            tracked the state correctly.
        """
        if chat_id in self.message_with_inline_keyboard:
            message_id = self.message_with_inline_keyboard[chat_id]
            await self.bot.edit_message_reply_markup(chat_id, message_id, reply_markup=None)
            del self.message_with_inline_keyboard[chat_id]
        if chat_id in self.chat_state:
            del self.chat_state[chat_id]
            await self.bot.send_message(chat_id, "Action cancelled", reply_markup=aiogram.types.ReplyKeyboardRemove())
        elif not silent:
            await self.bot.send_message(chat_id, "No action in progress.",
                                        reply_markup=aiogram.types.ReplyKeyboardRemove())

    def generic_connector(self, type_: Type[T], name: str,
                          to_message: Callable[[T], str], parse_value: Callable[[str], T],
                          read_roles: Set[RoleT], set_roles: Optional[Set[RoleT]] = None,
                          send_users: Set[UserT] = set(), options: Optional[List[str]] = None
                          ) -> "TelegramConnector[T, RoleT]":
        """
        Create a new connector object with any type, using the given formatting and parsing callback functions

        :param type_: The value type of this connector. Will be used as the `.type` attribute, used for SHC's static
            type checking
        :param name: The connector's name, which users will use for searching and selecting the connector to read and
            write its value.
        :param to_message: A formatting function, used for converting a value into a string, when read by the user or
            sent to the user by subscription
        :param parse_value: A function for parsing values sent by the user for publishing in SHC
        :param read_roles: A set of user roles which are allowed to read the current value from the connector. The role
            objects must be valid authorization roles of the `auth_provider` of this bot interface.
        :param set_roles: A set of user roles which are allowed to send value updates via Telegram to the connector,
            i.e. to change the value of *connected* objects. The role objects must be valid authorization roles of the
            `auth_provider` of this bot interface.
        :param send_users: A set of users which shall receive a Telegram message for every value update published to the
            connector within SHC. The role objects must be valid user objects of the `auth_provider` of this bot
            interface.
        :param options: A list of values the user can select from for sending a new value. They are used to create a
            custom keyboard, which is shows to the user when selecting this connector. The options must be strings,
            which are supported to be parsed by the `parse_value` function. If this parameter is omitted, no custom
            keyboard is shown, i.e. the user is expected to use the normal text keyboard.
        """
        if name in self.connectors:
            raise ValueError(f"Connector with name {name} already exists in this Telegram bot.")
        var = TelegramConnector(
            self, type_, name, read_roles, send_users,
            set_roles if set_roles is not None else read_roles,
            "Change to?", parse_value,
            lambda x: f"{name} is currently {to_message(x)}", lambda x: f"{name} is now {to_message(x)}",
            options)
        self.connectors[name] = var
        return var

    def str_connector(self, name: str, read_roles: Set[RoleT], set_roles: Optional[Set[RoleT]] = None,
                      send_users: Set[UserT] = set()) -> "TelegramConnector[str, RoleT]":
        """
        Create a new connector object with string type values

        :param name: The connector's name, which users will use for searching and selecting the connector to read and
            write its value.
        :param read_roles: A set of user roles which are allowed to read the current value from the connector. The role
            objects must be valid authorization roles of the `auth_provider` of this bot interface.
        :param set_roles: A set of user roles which are allowed to send value updates via Telegram to the connector,
            i.e. to change the value of *connected* objects. The role objects must be valid authorization roles of the
            `auth_provider` of this bot interface.
        :param send_users: A set of users which shall receive a Telegram message for every value update published to the
            connector within SHC. The role objects must be valid user objects of the `auth_provider` of this bot
            interface.
        """
        if name in self.connectors:
            raise ValueError(f"Connector with name {name} already exists in this Telegram bot.")
        var = TelegramConnector(
            self, str, name, read_roles, send_users,
            set_roles if set_roles is not None else read_roles,
            "Change to?", lambda x: x,
            lambda x: f"{name} is currently {x}", lambda x: f"{name} is now {x}",
            None)
        self.connectors[name] = var
        return var

    def on_off_connector(self, name: str, read_roles: Set[RoleT], set_roles: Optional[Set[RoleT]] = None,
                         send_users: Set[UserT] = set()) -> "TelegramConnector[bool, RoleT]":
        """
        Create a new connector object with bool values, represented to the user as 'on' and 'off'

        :param name: The connector's name, which users will use for searching and selecting the connector to read and
            write its value.
        :param read_roles: A set of user roles which are allowed to read the current value from the connector. The role
            objects must be valid authorization roles of the `auth_provider` of this bot interface.
        :param set_roles: A set of user roles which are allowed to send value updates via Telegram to the connector,
            i.e. to change the value of *connected* objects. The role objects must be valid authorization roles of the
            `auth_provider` of this bot interface.
        :param send_users: A set of users which shall receive a Telegram message for every value update published to the
            connector within SHC. The role objects must be valid user objects of the `auth_provider` of this bot
            interface.
        """
        if name in self.connectors:
            raise ValueError(f"Connector with name {name} already exists in this Telegram bot.")

        def parse_value(x: str) -> bool:
            if x == "on":
                return True
            elif x == "off":
                return False
            else:
                raise ValueError("Invalid on/off value.")

        var = TelegramConnector(
            self, bool, name, read_roles, send_users,
            set_roles if set_roles is not None else read_roles,
            "Switch?",
            parse_value,
            lambda x: f"{name} is currently {'on' if x else 'off'}.",
            lambda x: f"{name} is now {'on' if x else 'off'}",
            ['off', 'on'])
        self.connectors[name] = var
        return var

    def trigger_connector(self, name: str, read_roles: Set[RoleT], set_roles: Optional[Set[RoleT]] = None,
                          send_users: Set[UserT] = set()) -> "TelegramConnector[None, RoleT]":
        """
        Create a new connector object None-type values

        This connector type is primarily meant to trigger handler functions. When a user with `set` authorization
        selects such a connector, they will be asked "Trigger?" and presented with a custom keyboard with a "do" and a
        "/cancel" button. Sending "do" will make the connector publish a None value (which is valid for triggering
        handlers via :meth:`.trigger() <TelegramConnector.trigger>`).

        The connector can also be used to send info messages to users: When :meth:`.write() <TelegramConnector.write>`
        is called, it will send a message "{name} has been triggered" to all users in the `send_users` set. However, the
        connector does not support reading values.

        :param name: The connector's name, which users will use for searching and selecting the connector to read and
            write its value.
        :param read_roles: A set of user roles which are allowed to read the current value from the connector. The role
            objects must be valid authorization roles of the `auth_provider` of this bot interface.
        :param set_roles: A set of user roles which are allowed to send value updates via Telegram to the connector,
            i.e. to change the value of *connected* objects. The role objects must be valid authorization roles of the
            `auth_provider` of this bot interface.
        :param send_users: A set of users which shall receive a Telegram message for every value update published to the
            connector within SHC. The role objects must be valid user objects of the `auth_provider` of this bot
            interface.
        """
        if name in self.connectors:
            raise ValueError(f"Connector with name {name} already exists in this Telegram bot.")

        def parse_value(x: str) -> None:
            if x == "do":
                return None
            else:
                raise ValueError("Invalid value. Must be 'do'. Otherwise, use /cancel to cancel triggering.")

        var = TelegramConnector(
            self, type(None), name, read_roles, send_users,
            set_roles if set_roles is not None else read_roles,
            "Trigger?",
            parse_value,
            lambda x: "",
            lambda x: f"{name} has been triggered",
            ['do'])
        self.connectors[name] = var
        return var

    async def send_message(self, text: str, users: Set[UserT], chat_ids: Set[int] = set()) -> None:
        """
        Send a message to one or more Telegram chats, identified by either the Telegram chat id or a user of the given
        auth_provider

        Currently, only unformatted text messages are supported.

        :param text: The message text
        :param users: A set of users of this bot's auth_provider to send the message to
        :param chat_ids: A set of (additional) telegram chat ids to send the message to. If a specific chat is part of
            both sets, the message is still only sent once.
        """
        user_chats = set()
        for user in users:
            chat = self.auth_provider.get_telegram_chat_of_user(user)
            if chat is None:
                logger.warning("Could not resolve user %s to Telegram chat_id for sending message to them", user)
            else:
                user_chats.add(chat)
        await asyncio.gather(*(self.bot.send_message(chat_id, text)
                               for chat_id in chat_ids | user_chats))


class TelegramConnector(Generic[T, RoleT], Reading[T], Subscribable[T], Writable[T],
                        metaclass=abc.ABCMeta):
    is_reading_optional = False

    def __init__(self, interface: TelegramBot, type_: Type[T], name: str, read_roles: Set[RoleT],
                 send_users: Set[UserT], set_roles: Set[RoleT], set_message: str, parse_value: Callable[[str], T],
                 format_value_read: Callable[[T], str], format_value_send: Callable[[T], str],
                 options: Optional[List[str]]):
        self.type = type_
        super().__init__()
        self.interface = interface
        self.name = name
        self.read_roles = read_roles
        self.send_users = send_users
        self.set_roles = set_roles

        self.set_message = set_message
        self.parse_value_fn = parse_value
        self.format_value_read_fn = format_value_read
        self.format_value_send_fn = format_value_send
        if options:
            self.keyboard = aiogram.types.ReplyKeyboardMarkup(
                [[aiogram.types.KeyboardButton(o)
                  for o in options[i:i+2]]
                 for i in range(0, len(options), 2)])
        else:
            self.keyboard = None

    def is_settable(self) -> bool:
        """ Returns True, if the connector has subscribers, i.e. it is meaningfully readable from Telegram """
        return bool(self._subscribers or self._triggers)

    def is_readable(self) -> bool:
        """ Returns True, if the connector has a default_provider, i.e. it is meaningfully writable from Telegram """
        return self._default_provider is not None

    async def _write(self, value: T, _origin: List[Any]) -> None:
        await self.interface.send_message(text=self.format_value_send_fn(value), users=self.send_users)

    async def read_message(self) -> Optional[str]:
        """ Create a response to a read-request from Telegram.

        :return: A message with the current value from the default_provider, using the :attr:`format_value_read_fn` or
        """
        value = await self._from_provider()
        if value is not None:
            return self.format_value_read_fn(value)
        return None

    def from_telegram(self, value: str) -> None:
        self._publish(self.parse_value_fn(value), [])

    def get_setting_keyboard(self) -> aiogram.types.ReplyKeyboardMarkup:
        return self.keyboard

    def get_set_message(self) -> str:
        return self.set_message


class TelegramAuthProvider(Generic[UserT, RoleT], metaclass=abc.ABCMeta):
    """
    Abstract base class for authentication & authorization providers for the SHC TelegramBot interface

    A `TelegramAuthProvider` allows to authenticate Telegram users by their chat id and check their authorization for
    specific functions of the TelegramBot interface based on a custom 'Role' type. The AuthProvider's Role type is used
    to define the set of authorized roles for each individual function of the TelegramBot interface. The set of roles is
    treated as a black box by the bot and only passed to the AuthProvider to check authorization.

    When a user interacts with the bot, the bot authenticates them by their chat id using :meth:`get_telegram_user`. If
    the method returns `None`, the user is rejected. Otherwise, the authenticated user is returned, represented as an
    (arbitrary) unique user id string. For each interaction with a Bot functionality, the AuthProvider's
    :meth:`has_user_role` method is used to check if the specific user (user id string) is allowed to use the
    functionality based on the given set of role objects.

    Provided implementations by SHC:

    * :class:`SimpleTelegramAuth`
    """
    @abc.abstractmethod
    def get_telegram_user(self, chat_id: int) -> Optional[UserT]:
        pass

    @abc.abstractmethod
    def has_user_role(self, user: UserT, roles: Set[RoleT]) -> Optional[bool]:
        pass

    def get_telegram_chat_of_user(self, user: UserT) -> Optional[int]:
        return None


class SimpleTelegramAuth(TelegramAuthProvider[str, str]):
    """
    A simple implementation of a :class:`TelegramAuthProvider`.

    It uses a fixed dict for mapping user id strings and telegram chat ids and uses the user id strings themselves as
    role definitions. Thus, it does not have a concept of user groups; you always need to specify the set of all
    authorized user ids for each TelegramBot connector. (Of course, you can put the set in a variable to reuse it.)

    :param users: Dict of users in the form {user_id: telegram_chat_id}
    """
    def __init__(self, users: Dict[str, int]):
        self.users = users
        self.user_by_id = {v: k for k, v in users.items()}

    def get_telegram_user(self, chat_id: int) -> Optional[str]:
        try:
            return self.user_by_id[chat_id]
        except KeyError:
            return None

    def has_user_role(self, user: str, roles: Set[RoleT]) -> Optional[bool]:
        return user in roles

    def get_telegram_chat_of_user(self, user: str) -> Optional[int]:
        try:
            return self.users[user]
        except KeyError:
            return None
